/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER
/// and THE UNIVERSITY OF MANCHESTER, seanb@cs.man.ac.uk

////////////////////////////////////////////////////////////////////////////
/// FaCT Client                                                          ///
/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER                     ///
/// and THE UNIVERSITY OF MANCHESTER                                     /// 
///					                                 ///
/// This program is free software; you can redistribute it and/or        ///
/// modify it under the terms of the GNU General Public License          ///
/// as published by the Free Software Foundation; either version 2       ///
/// of the License, or (at your option) any later version.               ///
///                                                                      ///
/// This program is distributed in the hope that it will be useful,      ///
/// but WITHOUT ANY WARRANTY; without even the implied warranty of       ///
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        ///
/// GNU General Public License for more details.                         ///
///                                                                      ///
/// You should have received a copy of the GNU General Public License    ///
/// along with this program; if not, write to the Free Software          ///
/// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            ///
/// 02111-1307, USA.                                                     ///
///                                                                      ///
/// Enquiries about FaCT should be directed to:                          ///
///                                                                      ///
/// email: horrocks@cs.man.ac.uk                                         ///
/// www:   http://www.cs.man.ac.uk/~horrocks                             ///
/// smail: Ian Horrocks, Department of Computer Science, Oxford Road,    ///
///       Manchester 467M13 9PL, United Kingdom                          ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
///                                                                      ///
/// This class is intended to demonstrate how to write a Java client     ///
/// that will communicate with the FaCT description logic reasoner via   ///
/// the CORBA-based interface. It is not supported or maintained and no  ///
/// guarantees are supplied as to the quality or robustness of the code. ///
/// SKB, 4/10/99                                                         ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

package img.fact.client;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.border.*;
import javax.swing.table.*;
import java.net.*;
import lispwrapper.*;
import org.omg.CORBA.*;
import org.omg.CosNaming.*;
import img.fact.cexpr.*;
import img.fact.*;
import img.util.*;

public class ClientApplet extends BasicClient {

public void actionPerformed(ActionEvent e) {
  // Connect; Disconnect; Quit; Begin; End; Reset; Abort
  String cmd = e.getActionCommand();
  if (cmd.startsWith("KB: ")) {
    this.readAndCompileKB(cmd);
  } else {
    super.actionPerformed(e);
  }
}

void addConnectionMenu(JMenuBar menuBar) {
  connMenu = new JMenu("Connection");
  connMenu.setMnemonic('c');

  JMenuItem item = new JMenuItem("Connect");
  item.setMnemonic('c');
  item.addActionListener(this);
  connMenu.add(item);

  item = new JMenuItem("Disconnect");
  item.setMnemonic('d');
  item.addActionListener(this);
  connMenu.add(item);

  menuBar.add(connMenu);
}

void addTellMenu(JMenuBar menuBar) {
  tellMenu = new JMenu("Tell");
  tellMenu.setMnemonic('t');

  JMenuItem item = new JMenuItem("Clear KB");
  item.setMnemonic('c');
  item.addActionListener(this);
  tellMenu.add(item);

  item = new JMenuItem("New Concept");
  item.setMnemonic('n');
  item.addActionListener(this);
  tellMenu.add(item);

  item = new JMenuItem("New Primitive");
  item.setMnemonic('p');
  item.addActionListener(this);
  tellMenu.add(item);
  addKBs();

  menuBar.add(tellMenu);
}

private void addKBs() {
  // Get a list of knowledge bases from the page and add items
  // to the menu accordingly.
  String num;
  int kbCount = -1;
  String names[];
  num = getParameter("kbs");
  try {
    kbCount = Integer.parseInt(num);
  } catch (NumberFormatException ex) {
  }
  for (int i = 0; i<kbCount; i++) {
    JMenuItem item = new JMenuItem("KB: " + getParameter("kb_" + Integer.toString(i)));
    item.addActionListener(this);
    tellMenu.add(item);
  }
}

  class CompilerThread extends Thread {
    boolean notDone = true;
    BasicClient lc;
    URL url;
    
    CompilerThread(BasicClient c, URL u) {
      lc = c;
      url = u;
    }
    
  public void run() {
    lc.compiling(true);
    lc.disableMenus();
    try {
      Compiler c = new Compiler(ch);
      c.setLogger(lc);
      c.clearBeforeCompiling(false);
      c.compile(url);
      lc.updateTells();
    } catch (Compiler.CompilerException ex) {
      lc.report("Compilation Problem: " + ex.information);
    }
    lc.enableMenus();
    lc.compiling(false);
  }
  }

private void readAndCompileKB(String cmd) {
  String nm = cmd.substring(4);
  String kb = getParameter(nm);
  if (kb!=null) {
    URL docBase = getDocumentBase();
    URL kbURL = null;
    try {
      kbURL = new URL(docBase, kb);
    } catch (MalformedURLException e) {
      System.out.println(e.getMessage());
    }
    CompilerThread ct = new CompilerThread(this, kbURL);
    ct.start();
  }
}
}
