/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER
/// and THE UNIVERSITY OF MANCHESTER, seanb@cs.man.ac.uk

////////////////////////////////////////////////////////////////////////////
/// FaCT Client                                                          ///
/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER                     ///
/// and THE UNIVERSITY OF MANCHESTER                                     /// 
///					                                 ///
/// This program is free software; you can redistribute it and/or        ///
/// modify it under the terms of the GNU General Public License          ///
/// as published by the Free Software Foundation; either version 2       ///
/// of the License, or (at your option) any later version.               ///
///                                                                      ///
/// This program is distributed in the hope that it will be useful,      ///
/// but WITHOUT ANY WARRANTY; without even the implied warranty of       ///
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        ///
/// GNU General Public License for more details.                         ///
///                                                                      ///
/// You should have received a copy of the GNU General Public License    ///
/// along with this program; if not, write to the Free Software          ///
/// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            ///
/// 02111-1307, USA.                                                     ///
///                                                                      ///
/// Enquiries about FaCT should be directed to:                          ///
///                                                                      ///
/// email: horrocks@cs.man.ac.uk                                         ///
/// www:   http://www.cs.man.ac.uk/~horrocks                             ///
/// smail: Ian Horrocks, Department of Computer Science, Oxford Road,    ///
///       Manchester 467M13 9PL, United Kingdom                          ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
///                                                                      ///
/// This class is intended to demonstrate how to write a Java client     ///
/// that will communicate with the FaCT description logic reasoner via   ///
/// the CORBA-based interface. It is not supported or maintained and no  ///
/// guarantees are supplied as to the quality or robustness of the code. ///
/// SKB, 4/10/99                                                         ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

package img.fact.client;

import java.util.*;
import java.io.*;
import java.net.*;
import lispwrapper.*;
import org.omg.CORBA.*;
import img.fact.*;
import img.fact.xml.*;
import EDU.auburn.VGJ.graph.*;
import EDU.auburn.VGJ.gui.*;
import EDU.auburn.VGJ.util.*;

/** Simple class that produces a graph and then displays it using the
    VGJ graph displayer */
public class Grapher {

private SimpleGraphWindow sgw;
private Graph graph;
private Hashtable nodes;
private ClientHandler ch; 
private BasicClient client;
private XMLToLisp xmlConverter;

  /** Constructor 
   * @param handler A client handler that'll be used to
   * communicate with the server. */
public Grapher(BasicClient bc, ClientHandler handler) {
  client = bc;
  ch = handler;
  graph = new Graph(true);
  try {
    xmlConverter = new XMLToLisp();
    //    xmlConverter.setPretty(true);
  } catch (XMLConversionException e) {
    System.err.println("Grapher Error: " + e.getMessage());
  }
}

private void resetGraph() {
  Graph tmpgraph = new Graph(true);
  graph.copy(tmpgraph);
  nodes = new Hashtable();
  sgw = new SimpleGraphWindow();
  sgw.pack();
  sgw.show();
}

  /** Show the hierarchy underneath the given concept */

public void showHierarchy(String con, int depth) {
  resetGraph();
  addNode(con);
  addEdges(con,depth);
  sgw.setGraph(graph);
}

public void showHierarchy(String d, taxonomyNode t, int depth) {
  resetGraph();
  String ld = "";
  try {
    ld = xmlConverter.convert(d);
  } catch (XMLConversionException ex) {
    ld = d;
  }
  addRecNode(ld);
  String [][] parents = t.supers;
  String [][] children = t.subs;
  try {
    for(int i=0; i < parents.length;i++) {
      String par = parents[i][0];
      String normalPar = xmlConverter.convert(par);
      addNode(normalPar);
      addEdgesUp(normalPar);
      addEdge(normalPar,ld);
    }
    for(int i=0; i < children.length;i++) {
      String kid = children[i][0];
      String normalKid = xmlConverter.convert(kid);
      addNode(normalKid);
      addEdgesDown(normalKid, depth-1);
      addEdge(ld,normalKid);
    }
  } catch (XMLConversionException e) {
    System.err.println("Grapher Error: " + e.getMessage());
  }
  sgw.setGraph(graph);
}

private void addEdge(String n1, String n2) {
  int nod1 = ((Integer) nodes.get(n1)).intValue();
  int nod2 = ((Integer) nodes.get(n2)).intValue();
  graph.insertEdge(nod1,nod2);
}

private boolean addNode(String con) {
  if (!nodes.containsKey(con)) {
    int index = graph.insertNode();
    Node nod = graph.getNodeFromIndex(index);
    nod.setLabel(con);
    nod.setShape(Node.OVAL);
    nod.setLabelPosition("In");
    // nod.setBoundingBox(40,40);
    nodes.put(con,new Integer(index));
    return true;
  } else {
    return false;
  } 
}

private boolean addRecNode(String con) {
  if (!nodes.containsKey(con)) {
    int index = graph.insertNode();
    Node nod = graph.getNodeFromIndex(index);
    nod.setLabel(con);
    nod.setShape(Node.RECTANGLE);
    nod.setLabelPosition("In");
    // nod.setBoundingBox(40,40);
    nodes.put(con,new Integer(index));
    return true;
  } else {
    return false;
  } 
}

private void addEdges(String con, int depth) {
  addEdgesDown(con, depth);
  addEdgesUp(con);
}

private void addEdgesDown(String con, int depth) {
  Vector results = null;
  if (con.length()!=0 && (depth > 0)) {
    try {
      String xmlCon = "";
      // First need to convert the expression to XML.
      if (con.equals(":TOP")) {
	xmlCon = "<TOP/>";
      } else if (con.equals(":BOTTOM")) {
	xmlCon = "<BOTTOM/>";
      } else {
	xmlCon = "<PRIMITIVE NAME=\"" + con + "\"/>";
      }
      String[][] res = ch.directSubsC(xmlCon);
      for (int i=0; i< res.length; i++) {
	String kid = res[i][0];
	// Now we need to convert the results back from XML format
	// to being names.
	String normalKid = xmlConverter.convert(kid);
	if (!normalKid.equals(":BOTTOM")) {
	  boolean recurse = addNode(normalKid);
	  int nod1 = ((Integer) nodes.get(con)).intValue();
	  int nod2 = ((Integer) nodes.get(normalKid)).intValue();
	  graph.insertEdge(nod1, nod2);
	  if (recurse) {
	    addEdgesDown(normalKid, depth-1);
	  }
	}
      } 
    } catch (XMLConversionException ex) {
      System.err.println("XML Error in Grapher: " + ex.getMessage());
    } catch (InternalException ex) {
      System.err.println(Short.toString(ex.code) + " " + ex.information);
    } catch (KBModifiedException ex) {
      System.err.println(Short.toString(ex.code) + " " + ex.information);
    } catch (OpUnimplementedException ex) {
      System.err.println(Short.toString(ex.code) + " " + ex.information);
    } catch (ExprErrorException ex) {
      System.err.println(Short.toString(ex.code) + " " + ex.information);
    }
  }
}
  
private void addEdgesUp(String con) {
  Vector results = null;
  if (con.length()!=0) {
    try {
      String xmlCon = "";
      // First need to convert the expression to XML.
      if (con.equals(":TOP")) {
	xmlCon = "<TOP/>";
      } else if (con.equals(":BOTTOM")) {
	xmlCon = "<BOTTOM/>";
      } else {
	xmlCon = "<PRIMITIVE NAME=\"" + con + "\"/>";
      }
      String[][] res = ch.directSupersC(xmlCon);
      for (int i=0; i< res.length; i++) {
	String par = res[i][0];
	// Now we need to convert the results back from XML format
	// to being names.
	String normalPar = xmlConverter.convert(par);
	boolean recurse = addNode(normalPar);
	int nod1 = ((Integer) nodes.get(con)).intValue();
	int nod2 = ((Integer) nodes.get(normalPar)).intValue();
	graph.insertEdge(nod2, nod1);
	if (recurse) {
	  addEdgesUp(normalPar);
	}
      }
    } catch (XMLConversionException ex) {
      System.err.println("XML Error in Grapher: " + ex.getMessage());
    } catch (InternalException ex) {
      System.err.println(Short.toString(ex.code) + " " + ex.information);
    } catch (KBModifiedException ex) {
      System.err.println(Short.toString(ex.code) + " " + ex.information);
    } catch (OpUnimplementedException ex) {
      System.err.println(Short.toString(ex.code) + " " + ex.information);
    } catch (ExprErrorException ex) {
      System.err.println(Short.toString(ex.code) + " " + ex.information);
    }

  }
}


}
