/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER
/// and THE UNIVERSITY OF MANCHESTER, seanb@cs.man.ac.uk

////////////////////////////////////////////////////////////////////////////
/// FaCT Client                                                          ///
/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER                     ///
/// and THE UNIVERSITY OF MANCHESTER                                     /// 
///					                                 ///
/// This program is free software; you can redistribute it and/or        ///
/// modify it under the terms of the GNU General Public License          ///
/// as published by the Free Software Foundation; either version 2       ///
/// of the License, or (at your option) any later version.               ///
///                                                                      ///
/// This program is distributed in the hope that it will be useful,      ///
/// but WITHOUT ANY WARRANTY; without even the implied warranty of       ///
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        ///
/// GNU General Public License for more details.                         ///
///                                                                      ///
/// You should have received a copy of the GNU General Public License    ///
/// along with this program; if not, write to the Free Software          ///
/// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            ///
/// 02111-1307, USA.                                                     ///
///                                                                      ///
/// Enquiries about FaCT should be directed to:                          ///
///                                                                      ///
/// email: horrocks@cs.man.ac.uk                                         ///
/// www:   http://www.cs.man.ac.uk/~horrocks                             ///
/// smail: Ian Horrocks, Department of Computer Science, Oxford Road,    ///
///       Manchester 467M13 9PL, United Kingdom                          ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
///                                                                      ///
/// This class is intended to demonstrate how to write a Java client     ///
/// that will communicate with the FaCT description logic reasoner via   ///
/// the CORBA-based interface. It is not supported or maintained and no  ///
/// guarantees are supplied as to the quality or robustness of the code. ///
/// SKB, 4/10/99                                                         ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

package img.fact.client;

/** This is used to perform long operation on a seperate thread, such that
 *  the Java UI thread will not be blocked and windows smudged.
 *  
 *  The user/caller of this thread may want to know that this thread is indeed
 *  going to take a long time. "A long time" is subjective. For this reason, 
 *  a timeout can be associated with this thread, when timeout has elapsed, 
 *  onTimeOut() will be called. This feature can be disabled by using 
 *  negative timeout in the constructor. 
 *  
 *  The run() method will first launch another thread to do the timing,
 *  then call performOperation() and finally doneOperation(). 
 *  If performOperation() finishes before 
 *  timeout has expired, the timing is interrupted and therefore 
 *  onTimeOut will never be called (i.e. a short operation).
 *
 * To use this, just simply create an instance of this and override the 
 * following functions inline:
 * <UL>
 * <LI>performOperation</LI>
 * <LI>onTimeOut</LI>
 * <LI>doneOperation</LI>
 * </UL>
 * 
 * Note: an alternative implementation: a calling class implementing 
 * an interface rather than inline overridding. BUT overridding allows 
 * this to be used at multiple points in a class, each doing a different 
 * operation, whereas asking that class to implement an interface 
 * would restrict the number of long operations per class to ONE. 
 * 
 * @author Gary Ng
 * @version Jan 13, 1999 (v0.83)
 *
 */
class WaitThread extends Thread {
  boolean notDone = true;

  /** to be overloaded by whichever class uses this thread 
   *  this is where the operations to be threaded go.
   */
  public void performOperation() {
  }

  /** to be overloaded by whichever class uses this thread 
   *  this is called when the thread's run() is about to finish
   */
  public void doneOperation() {
  }

  public void run() {
    performOperation();
    // The following works fine as application but applet cannot modify a 
    // thread due to security!!!. so this has to be checked by a boolean.
    //    timing.interrupt(); 
    notDone = false;
    //    System.out.println("done");
    doneOperation();
  }
}
